import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.util.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;

/**
 * <h1>SoftKeyboardExperiment</h1>
 * 
 * <h3>Summary</h3>
 * 
 * <ul>
 * <li>Experiment software for text entry using a soft keyboard.
 * <p>
 * 
 * <li>Several soft keyboard layouts are supported:
 * <p>
 * 
 * <ul>
 * <li>Qwerty
 * <li>Opti
 * <li>Metropolis
 * <li>LKL (Lewis-Kennedy-LaLomia)
 * <li>Fitaly
 * <li>Custom (user defined)
 * </ul>
 * </ul>
 * 
 * <h3>Related References</h3>
 * 
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/CHI99a.html">The design and evaluation of a high performance
 * soft keyboard</a>, by MacKenzie and Zhang (<i>CHI '99</i>). This is the paper where the Opti
 * design was introduced. Opti was the first soft keyboard to be designed using a Fitts-digram
 * model.
 * <p>
 * 
 * <li>
 * <a href="http://dl.acm.org/citation.cfm?id=633380">Physics-based graphical keyboard design</a>,
 * by Hunter, Zhai, and Smith (<i>CHI 2000</i>). This paper introduced the Metropolis soft keyboard,
 * offered as an alternative Opti. Metropolis was also designed using a Fitts-digram model; however,
 * the design process was automated using an algorithm.
 * <p>
 * 
 * <li>
 * <a href="http://pro.sagepub.com/content/43/5/420.short">Evaluation of typing key layouts for
 * stylus input</a>, by Lewis, LaLomia, & Kennedy (<i>HFES '99</i>). This is the paper introducing
 * the LKL (Lewis-Kennedy-LaLomia) keyboard.
 * <p>
 * 
 * <li>
 * <a href="http://www.fitaly.com/">Textware Solutions</a> web site. Textware Solutions is the
 * company that introduced the Fitaly soft keyboard.
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/hci3.html">Text entry for mobile computing: Models and methods,
 * theory and practice</a>, by MacKenzie and Soukoreff (<i>Human-Computer Interaction</i>, 2002).
 * This paper includes an extensive section on soft keyboards, including a comparison of entry speed
 * predictions using a Fitts-digram model. The comparison includes Qwerty, Opti, Metropolis, LKL,
 * Fitaly, and several other layouts (Figure 21).
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/bit95.html">Theoretical upper and lower bounds on typing speed
 * using a stylus and soft keyboard</a>, by Soukoreff and MacKenzie (<i>Behaviour & Information
 * Technology</i>, 1995). This is the paper that introduced the Fitts-digram model for designing
 * soft keyboards.
 * </ul>
 * 
 * <h3>Background</h3>
 * 
 * The goal in designing Opti (and other layouts) was to minimize stylus or finger movement for
 * English text entry on a soft keyboard. As an example, consider the movements required to enter
 * "airplane":
 * <p>
 * 
 * <center> <img src="SoftKeyboardExperiment-5.jpg"> </center>
 * <p>
 * 
 * Clearly, there is less finger or stylus movement for the Opti design. The trick is to minimize
 * the movement as an overall average for text entry in a particular language, such as English. See
 * the papers above for details.
 * <p>
 * 
 * <h3>Running the Experiment Software</h3>
 * <p>
 * 
 * <a href="http://www.yorku.ca/mack/HCIbook/Running/">Click here</a> for instructions on
 * launching/running the application.
 * <p>
 * 
 * <h3>Setup Parameters</h3>
 * 
 * Upon launching, the program presents a setup dialog:
 * <p>
 * 
 * <center> <img src="SoftKeyboardExperiment-1.jpg"> </center>
 * <p>
 * 
 * The default parameter settings are read from a configuration file called
 * <code>SoftKeyboardExperiment.cfg</code>. This file is created automatically when the application
 * is launched for the first time. The default parameter settings may be changed through the setup
 * dialog. The setup parameters are as follows:
 * <p>
 * 
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" valign="top">
 * <tr bgcolor="#cccccc">
 * <th>Parameter
 * <th>Description
 * 
 * <tr valign="top">
 * <td>Participant Code
 * <td>Identifies the current participant. This is used in forming the names for the output data
 * files. Also, the sd2 output data file includes a column with the participant code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Condition code
 * <td>An arbitrary code used to associate a test condition with this invocation. This parameter
 * might be useful if the software is used in an experiment, since conditions might be tested that
 * are not inherently part of the application (e.g., Gender &rarr; male, female; Input method &rarr;
 * mouse, stylus). The condition code is used in forming the name for the output data file. Also,
 * the sd2 output data file contains a column with the condition code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Session code
 * <td>Identifies the session. Useful if testing proceeds over multiple sessions to gauge the
 * progression of learning. The session code is used in forming the name for the output data file.
 * Also, the sd2 output data file contains a column with the condition code.
 * <p>
 * 
 * Note: The setup dialog does not include an entry for "Block code". The block code is generated
 * automatically by the software.
 * <p>
 * 
 * <tr valign="top">
 * <td>Number of phrases
 * <td>Specifies the number of phrases presented to the participant in the current block.
 * <p>
 * 
 * <tr valign="top">
 * <td>Phrases file
 * <td>Specifies a file containing phrases of text to be presented to participants for entry.
 * Phrases are drawn from the file at random. Typically, <a href="phrases2.txt">phrases2.txt</a> is
 * used. This is the phrase set published by MacKenzie and Soukoreff in 2003 (see Related
 * references, above). Other file names may be specified. For example, the file <a
 * href="quickbrownfox.txt">quickbrownfox.txt</a> contains the phrase
 * <p>
 * 
 * <pre>
*     the quick brown fox jumps over the lazy dog
* </pre>
 * 
 * This phrase might be useful for demonstration or specialized testing.
 * <p>
 * 
 * <tr valign="top">
 * <td>Show presented text during input
 * <td>A checkbox item that determines whether or not the presented text is visible during entry of
 * a phrase. Either way, the text phrase appears at the beginning of a trial. If this option is
 * unchecked, the phrase will disappear when the first keystroke is entered.
 * <p>
 * 
 * <tr valign="top">
 * <td>Keyboard
 * <td>Identifies the soft keyboard to use in the experiment. The options currently supported are as
 * follows: (click to enlarge)
 * <p>
 * 
 * <center> <b>Qwerty</b><br>
 * <a href="SoftKeyboardExperiment-qwerty.jpg"><img src="SoftKeyboardExperiment-qwerty.jpg"
 * width=150></a>
 * <p>
 * 
 * <b>Opti</b><br>
 * <a href="SoftKeyboardExperiment-opti.jpg"><img src="SoftKeyboardExperiment-opti.jpg"
 * width=150></a>
 * <p>
 * 
 * <b>Metropolis</b><br>
 * <a href="SoftKeyboardExperiment-metropolis.jpg"><img src="SoftKeyboardExperiment-metropolis.jpg"
 * width=150></a>
 * <p>
 * 
 * <b>Lewis-Kennedy-LaLommia</b><br>
 * <a href="SoftKeyboardExperiment-lewis.jpg"><img src="SoftKeyboardExperiment-lewis.jpg"
 * width=150></a>
 * <p>
 * 
 * <b>Fitaly</b><br>
 * <a href="SoftKeyboardExperiment-fitaly.jpg"><img src="SoftKeyboardExperiment-fitaly.jpg"
 * width=150></a>
 * <p>
 * 
 * <b>Custom</b><br>
 * (see below)
 * <p>
 * </center>
 * 
 * <tr valign="top">
 * <td>Custom keyboard
 * <td>Used in conjunction with the Custom option above to select a custom keyboard layout. The
 * layout is read from a keyboard definition file which is selected from a file chooser. The file
 * chooser opens when the "Browse..." button is clicked.
 * </table>
 * </blockquote>
 * <p>
 * 
 * <h3>Program Operation</h3>
 * 
 * The experiment launches with the selected soft keyboard displayed and with a text phrase
 * presented for input. An example is shown below for the Qwerty layout.
 * <p>
 * 
 * <center> <img src="SoftKeyboardExperiment-2.jpg"> </center>
 * <p>
 * 
 * At the end of each phrase, a popup dialog gives results for the phrase. Here's an example:
 * <p>
 * 
 * <center> <img src="SoftKeyboardExperiment-4.jpg"> </center>
 * <p>
 * 
 * <p>
 * 
 * <h3>Custom Mode</h3>
 * <p>
 * 
 * The last radio button in the setup dialog is "Custom (select below)". This option allows the
 * software to use a custom layout. The layout is read from a text file. The "Browse..." button pops
 * up a file chooser to select the keyboard definition file for the custom layout. An example is
 * provided for the well-known Dvorak layout: <code><a href="dvorak.txt">dvorak.txt</a></code>. If
 * this file is selected, the program will launch with the following soft keyboard:
 * <p>
 * 
 * <center><img src="SoftKeyboardExperiment-8.jpg"></center>
 * <p>
 * 
 * <h3>Output Data Files</h3>
 * 
 * At the end of a block of phrases, results are written to files. There are two output data files,
 * an "sd1" file and an "sd2" file. ("sd" is for "summary data".) The sd1 file contains data on a
 * keystroke by keystroke basis, including timestamps. The sd2 file contains summary data for each
 * phrase, one line per phrase. Examples follow:
 * <p>
 * 
 * <ul>
 * <li><a href="SoftKeyboardExperiment-sd1-example.txt">sd1 example</a>
 * <li><a href="SoftKeyboardExperiment-sd2-example.txt">sd2 example</a>
 * </ul>
 * 
 * Actual output files use "SoftKeyboardExperiment" as the base filename. This is followed by the
 * participant code, the condition code, the keyboard, the session code, and the block code. An
 * example might be <code>SoftKeyboardExperiment-P01-C01-OP-S01-B01.sd1</code>.
 * <p>
 * 
 * In most cases, the sd2 data files are the primary files used for data analyses in an experimental
 * evaluation. The data in the sd2 files are full-precision, comma-delimited, to facilitate
 * importing into a spreadsheet or statistics application. Below is an example for the sd2 file
 * above, after importing into Microsoft Excel: (click to enlarge)
 * 
 * <center>
 * <a href="SoftKeyboardExperiment-9.jpg"><img src="SoftKeyboardExperiment-9.jpg" width=800></a>
 * </center>
 * <p>
 * 
 * <h3>Miscellaneous</h3>
 * 
 * When using this program in an experiment, it is a good idea to terminate all other applications
 * and disable the system's network connection. This will maintain the integrity of the data
 * collected and ensure that the program runs without hesitations.
 * <p>
 * 
 * For demonstration, a mouse can be used for input, but finger or stylus input is more appropriate
 * and should be used if possible.
 * <p>
 * 
 * @author Scott MacKenzie, 2011-2015
 * @author Steven Castellucci, 2014
 */
public class SoftKeyboardExperiment
{
	final static String APP_NAME = "SoftKeyboardExperiment";
	final static int PARAMETERS = 8; // must equal the number of parameters defined below

	// setup parameters (1st value is default)
	final static String[] PARTICIPANT_CODES = { "P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07", "P08", "P09",
			"P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24",
			"P25", "P26", "P27", "P28", "P29", "P30" };
	final static String[] CONDITION_CODE = { "C00" };
	final static String[] SESSION_CODES = { "S00", "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09",
			"S10", "S11", "S12", "S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20", "S21", "S22", "S23", "S24",
			"S25", "S26", "S27", "S28", "S29", "S30" };
	final static String[] NUMBER_OF_PHRASES = { "5" };
	final static String[] PHRASES_FILE = { "phrases2.txt" };
	final static String[] SHOW_PRESENTED = { "true" };
	final static String[] KEYBOARD = { "Qwerty", "Opti", "Metropolis", "LKL (Lewis-Kennedy-LaLommia)", "Fitaly",
			"Custom (select below)" };
	final static String[] KEYBOARD_DEF_FILE = { null };

	// identify other files needed (used when executing from the .jar file for the 1st time)
	final static String[] OTHER_FILES = { "phrases2.txt", "helloworld.txt", "quickbrownfox.txt", "phrases.txt" };

	final static String[] KEYBOARD_CODE = { "QWERTY", "OPTI", "METROPOLIS", "LKL", "FITALY", "CUSTOM" };

	public static void main(String[] args) throws IOException
	{
		// use Win32 look and feel
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e)
		{
		}

		/*
		 * Define the default parameters settings. These are only used if the .cfg file does not
		 * exit, such as when the application is launched for the 1st time.
		 */
		String[] defaultParameters = new String[PARAMETERS];
		defaultParameters[0] = PARTICIPANT_CODES[0];
		defaultParameters[1] = CONDITION_CODE[0];
		defaultParameters[2] = SESSION_CODES[0];
		defaultParameters[3] = NUMBER_OF_PHRASES[0];
		defaultParameters[4] = PHRASES_FILE[0];
		defaultParameters[5] = SHOW_PRESENTED[0];
		defaultParameters[6] = KEYBOARD[0];
		defaultParameters[7] = KEYBOARD_DEF_FILE[0];

		/*
		 * Instantiate a Configuration object. The configuration parameters are (a) read from the
		 * .cfg file or (b) the passed default parameters if the .cfg file does not exist.
		 */
		Configuration c = Configuration.readConfigurationData(APP_NAME, defaultParameters, OTHER_FILES);
		if (c == null)
		{
			System.out.println("Error reading configuration data from " + APP_NAME + ".cfg!");
			System.exit(0);
		}

		/*
		 * Use the app's parameters to define the SetupItemInfo objects needed to create the setup
		 * dialog.
		 */
		SetupItemInfo[] sii = new SetupItemInfo[PARAMETERS];
		sii[0] = new SetupItemInfo(SetupItem.COMBO_BOX, "Participant code ", PARTICIPANT_CODES);
		sii[1] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Condition code ", CONDITION_CODE);
		sii[2] = new SetupItemInfo(SetupItem.COMBO_BOX, "Session code ", SESSION_CODES);
		sii[3] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Number of phrases ", NUMBER_OF_PHRASES);
		sii[4] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Phrases file ", PHRASES_FILE);
		sii[5] = new SetupItemInfo(SetupItem.CHECK_BOX, "Show presented text during input ", SHOW_PRESENTED);
		sii[6] = new SetupItemInfo(SetupItem.RADIO_BUTTON, "Keyboard ", KEYBOARD);
		sii[7] = new SetupItemInfo(SetupItem.FILE_CHOOSER, "Custom keyboard ", KEYBOARD_DEF_FILE);

		/*
		 * Create and show the setup dialog. The user can make changes to the default or existing
		 * configuration parameters
		 */
		Setup s = new Setup(null, c, APP_NAME, sii);
		s.showSetup(null);

		// launch application GUI
		SoftKeyboardExperimentGui screen = new SoftKeyboardExperimentGui(c);
		screen.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		screen.setTitle(APP_NAME);
		screen.pack();

		// put application in centre of screen
		int w = screen.getWidth();
		int h = screen.getHeight();
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getScreenSize();
		screen.setLocation((d.width - w) / 2, (d.height - h) / 2);
		screen.setVisible(true);
	}
}

class SoftKeyboardExperimentGui extends JFrame implements ActionListener
{
	// the following avoids a "warning" with Java 1.5.0 complier (?)
	static final long serialVersionUID = 42L;

	final String SD1_HEADER = "";
	final String SD2_HEADER = "App,Participant,Condition,Keyboard,Session,Block,"
			+ "Keystrokes,Characters,Time,MSD,Speed_(wpm),Error_rate_(%),KSPC\n";
	final Font F16 = new Font("sansserif", Font.PLAIN, 16);

	// mode
	final int QWERTY = 100;
	final int OPTI = 200;
	final int METROPOLIS = 300;
	final int LKL = 400;
	final int FITALY = 500;
	final int CUSTOM = 600;

	// define a mode array (values must be in same order as in setup dialog)
	final int[] MODE_ARRAY = { QWERTY, OPTI, METROPOLIS, LKL, FITALY, CUSTOM };

	private BufferedWriter sd1File;
	private BufferedWriter sd2File;
	private Vector<Sample> samples;
	private String phrasesFile;
	private boolean showPresented;
	private boolean newPhrase = true;
	private Random r = new Random();
	private JTextField text1;
	private JTextField text2;
	private SoftKeyboard sk;
	private String[] phrases;
	private String pPhrase; // presented phrase
	private String tPhrase; // transcribed phrase
	private long t1, t2;
	private int count; // count keystrokes per phrase
	private int phraseCount;
	private int keyboardInt;
	private String keyboardString;
	private String keyboardDefFile;
	private String appName;

	private String participantCode, conditionCode, sessionCode, blockCode, keyboard;

	private Font BIG = new Font("monospaced", Font.BOLD, 24);
	private Color BACKGROUND = new Color(254, 254, 218);
	private Color FOREGROUND = new Color(11, 11, 109);

	JDialog resultsDialog;
	JOptionPane resultsPane;
	JTextArea resultsArea;

	Configuration c;

	// constructor
	SoftKeyboardExperimentGui(Configuration cArg) throws IOException
	{
		// initialize setup parameters
		c = cArg;
		participantCode = c.getConfigurationParameter(0);
		conditionCode = c.getConfigurationParameter(1);
		sessionCode = c.getConfigurationParameter(2);
		// Note: Block code generated automatically by the software (see below)
		phraseCount = Integer.parseInt(c.getConfigurationParameter(3));
		phrasesFile = c.getConfigurationParameter(4);
		showPresented = Boolean.valueOf(c.getConfigurationParameter(5));
		keyboard = c.getConfigurationParameter(6);
		keyboardDefFile = c.getConfigurationParameter(7);

		// remap the keyboard parameter to useful integer and string variables
		int idx = -1;
		for (int i = 0; i < SoftKeyboardExperiment.KEYBOARD.length; ++i)
		{
			// System.out.println("i=" + i + ", keyboard=" + keyboard +
			// ", SoftKeyboardExperiment.KEYBOARD[i]=" + SoftKeyboardExperiment.KEYBOARD[i]);
			if (keyboard.equals(SoftKeyboardExperiment.KEYBOARD[i]))
			{
				idx = i;
				break;
			}
		}

		if (idx == -1)
		{
			System.out.println("Error: can't find keyboard: " + keyboard);
			System.exit(0);
		} else
		{
			keyboardInt = MODE_ARRAY[idx];
			keyboardString = SoftKeyboardExperiment.KEYBOARD_CODE[idx];
		}

		appName = SoftKeyboardExperiment.APP_NAME;

		// ------------------
		// initialize phrases
		// ------------------

		File f = new File(phrasesFile);
		if (!f.exists())
		{
			showError("\nPhrases file not found - " + phrasesFile);
			System.exit(1);
		}

		// open disk file for input
		BufferedReader inFile = new BufferedReader(new FileReader(phrasesFile));
		Vector<String> v = new Vector<String>();
		String s;
		while ((s = inFile.readLine()) != null)
			v.addElement(s.toLowerCase());
		phrases = new String[v.size()];
		v.copyInto(phrases);

		// -----------------------
		// initialize output files
		// -----------------------

		blockCode = "";
		int blockCodeNumber = 0;
		String base = SoftKeyboardExperiment.APP_NAME + "-" + participantCode + "-" + conditionCode + "-"
				+ keyboardString + "-" + sessionCode;

		String s1 = "";
		String s2 = "";
		do // find next available block code
		{
			++blockCodeNumber;
			blockCode = blockCodeNumber < 10 ? "B0" + blockCodeNumber : "B" + blockCodeNumber;
			s1 = base + "-" + blockCode + ".sd1";
			s2 = base + "-" + blockCode + ".sd2";
		} while (new File(s1).exists());

		try
		{
			sd1File = new BufferedWriter(new FileWriter(s1));
			sd2File = new BufferedWriter(new FileWriter(s2));
		} catch (IOException e)
		{
			System.out.println("I/O error: can't open sd1/sd2 data files");
			System.exit(0);
		}

		try
		{
			sd1File.write(SD1_HEADER, 0, SD1_HEADER.length());
			sd1File.flush();
			sd2File.write(SD2_HEADER, 0, SD2_HEADER.length());
			sd2File.flush();
		} catch (IOException e)
		{
			System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
			System.exit(1);
		}

		// initialize components
		text1 = new JTextField(50);
		text1.setFocusable(false);
		text1.setFont(BIG);
		text1.setBackground(new Color(254, 254, 218));
		text1.setForeground(new Color(11, 11, 109));
		JPanel text1Panel = new JPanel();
		text1Panel.setLayout(new FlowLayout(FlowLayout.LEFT));
		text1Panel.add(text1);
		text1Panel.setBorder(new TitledBorder(new EtchedBorder(), "Presented text"));

		text2 = new JTextField(50);
		text2.setEditable(true); // allows I-beam to show
		text2.setFont(BIG);
		text2.setBackground(BACKGROUND);
		text2.setForeground(FOREGROUND);
		JPanel text2Panel = new JPanel();
		text2Panel.setLayout(new FlowLayout(FlowLayout.LEFT));
		text2Panel.add(text2);
		text2Panel.setBorder(new TitledBorder(new EtchedBorder(), "Transcribed text"));

		// *** SOFT KEYBOARD ***
		if (keyboardInt == QWERTY)
			sk = new SoftKeyboard(qwertyKeyboard, this, keyboardInt);

		else if (keyboardInt == OPTI)
			sk = new SoftKeyboard(optiKeyboard, this, keyboardInt);

		else if (keyboardInt == METROPOLIS)
			sk = new SoftKeyboard(metropolisKeyboard, this, keyboardInt);

		else if (keyboardInt == LKL)
			sk = new SoftKeyboard(lewisKeyboard, this, keyboardInt);

		else if (keyboardInt == FITALY)
			sk = new SoftKeyboard(fitalyKeyboard, this, keyboardInt);

		else if (keyboardInt == CUSTOM)
		{
			Key[] customKeyboard = loadKeys(keyboardDefFile);
			sk = new SoftKeyboard(customKeyboard, this, keyboardInt);
		}

		JPanel keyboardPanel = new JPanel();
		keyboardPanel.add(sk);
		keyboardPanel.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));

		// present first phrase for input
		pPhrase = phrases[r.nextInt(phrases.length)];
		text1.setText(pPhrase);
		tPhrase = "";
		text2.setText(tPhrase);
		t1 = 0;
		count = 0;
		samples = new Vector<Sample>();

		// prepare results dialog
		resultsArea = new JTextArea(13, 40);
		resultsArea.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsArea.setBackground((new JButton()).getBackground());
		resultsPane = new JOptionPane(resultsArea, JOptionPane.INFORMATION_MESSAGE);
		resultsPane.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsDialog = resultsPane.createDialog(this, "Information");

		// layout components
		JPanel p1 = new JPanel();
		p1.setLayout(new BoxLayout(p1, BoxLayout.Y_AXIS));
		p1.add(Box.createRigidArea(new Dimension(5, 10)));
		p1.add(text1Panel);
		p1.add(Box.createRigidArea(new Dimension(5, 10)));
		p1.add(text2Panel);
		p1.add(Box.createRigidArea(new Dimension(5, 10)));

		JPanel p = new JPanel(new BorderLayout());
		p.add(p1, "North");
		p.add(keyboardPanel, "Center");
		this.setContentPane(p);
	}

	/*
	 * load a Key array from a custom keyboard definition file.
	 */
	Key[] loadKeys(String fileArg)
	{
		BufferedReader br = null;
		try
		{
			br = new BufferedReader(new FileReader(fileArg));
		} catch (FileNotFoundException e)
		{
			errorExit("File not found error!\nFile: " + fileArg);
		}

		/*
		 * The 1st data line contains the nominal width and height of keys. The x-position,
		 * y-position, width, and height for all the keys in the keyboard are computed relative to
		 * the nominal width and height.
		 */
		String line = getNextDataLine(br);
		String[] s = line.split(",");
		if (s.length != 2)
		{
			System.out.println("Error reading width/height!");
			System.exit(0);
		}
		int width = Integer.parseInt(s[0].trim());
		int height = Integer.parseInt(s[1].trim());

		/*
		 * The remaining data lines contain the definition of the keys in the custom keyboard.
		 */
		Vector<Key> v = new Vector<Key>();
		while ((line = getNextDataLine(br)) != null)
		{
			s = line.split(",");
			if (s.length != 5)
			{
				System.out.println("Error in key def!");
				System.exit(0);
			}
			String keyText = s[0].trim();
			int x = (int)(Double.parseDouble(s[1].trim()) * width + 0.5);
			int y = (int)(Double.parseDouble(s[2].trim()) * height + 0.5);
			int w = (int)(Double.parseDouble(s[3].trim()) * width + 0.5);
			int h = (int)(Double.parseDouble(s[4].trim()) * height + 0.5);
			v.add(new Key(keyText, x, y, w, h));
		}
		Key[] k = new Key[v.size()];
		v.copyInto(k);
		return k;
	}

	/*
	 * Read the next line from the keyboard definition file, skipping over blank lines or lines
	 * beginning with "#". If end-of-file is reached, return null.
	 */
	String getNextDataLine(BufferedReader brArg)
	{
		BufferedReader br = brArg;
		String s = null;
		do
		{
			try
			{
				s = br.readLine();
			} catch (IOException e)
			{
				System.out.println("IOException: " + e);
				System.exit(0);
			}
		} while (s != null && (s.length() == 0 || s.substring(0, 1).equals("#")));
		return s;
	}

	/*
	 * Output error message in a GUI popup (rather than printing to the console). Then exit.
	 */
	private void errorExit(String message)
	{
		JTextArea tmp = new JTextArea(message);
		tmp.setFont(F16);
		tmp.setBackground(new Color(212, 208, 200));
		JOptionPane.showMessageDialog(this, tmp, "Error", JOptionPane.ERROR_MESSAGE);
		System.exit(0);
	}

	void showError(String msg)
	{
		JOptionPane.showMessageDialog(null, msg, "I/O Error", JOptionPane.ERROR_MESSAGE);
	}

	public void actionPerformed(ActionEvent ae)
	{
		if (newPhrase)
		{
			newPhrase = false;
			if (!showPresented)
				text1.setText("");
		}

		JButton jb = (JButton)ae.getSource();
		String s = jb.getText();
		char c = (s.toLowerCase()).charAt(0);

		++count;
		if (t1 == 0)
			t1 = System.currentTimeMillis();
		t2 = System.currentTimeMillis() - t1;
		samples.addElement(new Sample(t2, s.toLowerCase()));

		// determine if Enter, Bksp, or "just a keystroke"
		if (s.equals("Enter"))
		{
			newPhrase = true;
			--count; // don't count pressing Enter as a keystroke

			String s1 = pPhrase.toLowerCase();
			String s2 = text2.getText().toLowerCase();

			MSD s1s2 = new MSD(s1, s2);
			int msd = s1s2.getMSD();

			// build formated results to output to GUI
			String results = " DATA COLLECTED:\n";
			results += String.format("   Presented:\t\t%s\n", s1);
			results += String.format("   Transcribed:\t%s\n", s2);
			results += String.format("   Keystrokes:\t\t%d\n", count);
			results += String.format("   Characters:\t\t%d (%.1f words)\n", s2.length(), s2.length() / 5.0);
			results += String.format("   Time:\t\t%.2f s (%.2f minutes)\n", t2 / 1000.0, t2 / 1000.0 / 60.0);
			results += String.format("   MSD:\t\t%d\n\n", msd);
			results += " PARTICIPANT PERFORMANCE:\n";
			results += String.format("   Entry Speed:\t%.2f wpm\n", wpm(s2, t2));
			results += String.format("   Error rate:\t\t%.2f%%\n", s1s2.getErrorRateNew());
			results += String.format("   KSPC:\t\t%.4f\n", (double)count / s2.length());
			resultsArea.setText(results);
			resultsDialog.setVisible(true);

			// build output data for sd2 file
			String sd2Stuff = appName + "," + participantCode + "," + conditionCode + "," + keyboardString + ","
					+ sessionCode + "," + blockCode + ",";
			sd2Stuff += String.format("%d,", count); // number of keystrokes
			sd2Stuff += String.format("%d,", s2.length()); // number of characters
			sd2Stuff += String.format("%f,", t2 / 1000.0); // time (in seconds)
			sd2Stuff += String.format("%d,", msd); // minimum string distance
			sd2Stuff += String.format("%f,", wpm(s2, t2)); // speed in words per minute
			sd2Stuff += String.format("%f,", s1s2.getErrorRateNew()); // error rate
			sd2Stuff += String.format("%f\n", (double)count / s2.length()); // KSPC

			// build output data for sd1 file
			String sd1Stuff = pPhrase + "\n" + s2 + "\n";
			for (int i = 0; i < samples.size(); ++i)
				sd1Stuff += samples.elementAt(i) + "\n";
			sd1Stuff += "#\n";

			// dump data
			try
			{
				sd1File.write(sd1Stuff, 0, sd1Stuff.length());
				sd1File.flush();
				sd2File.write(sd2Stuff, 0, sd2Stuff.length());
				sd2File.flush();
			} catch (IOException e)
			{
				System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
				System.exit(1);
			}

			// terminate if finished last phrase in block
			if (--phraseCount == 0)
			{
				JLabel thankyou = new JLabel("End of block. Thank you.");
				thankyou.setFont(new Font("sansserif", Font.PLAIN, 16));
				JOptionPane.showMessageDialog(this, thankyou);
				try
				{
					sd1File.close();
					sd2File.close();
				} catch (IOException e)
				{
					System.err.println("ERROR CLOSING DATA FILES!\n" + e);
					System.exit(1);
				}
				System.exit(0);
			}

			// prepare for next phrase
			samples = new Vector<Sample>();
			pPhrase = phrases[r.nextInt(phrases.length)];
			text1.setText(pPhrase);
			tPhrase = "";
			text2.setText(tPhrase);
			t1 = 0;
			count = 0;
		} else if (s.equals("Bksp"))
		{
			if (tPhrase.length() >= 1)
			{
				tPhrase = tPhrase.substring(0, tPhrase.length() - 1);
				text2.setText(tPhrase);
				text2.requestFocus(); // so I-beam (caret) does not disappear
			}
		} else
		// just a keystroke; add to transcribed text
		{
			tPhrase += c;
			text2.setText(tPhrase);
			text2.requestFocus(); // so I-beam (caret) does not disappear
		}
	} // end actionPerformed

	// compute typing speed in wpm given text entered and time in ms
	public static double wpm(String text, long msTime)
	{
		double speed = text.length();
		speed = speed / (msTime / 1000.0) * (60 / 5);
		return speed;
	}

	private class Sample
	{
		Sample(long l, String s)
		{
			time = l;
			key = s;
		}

		public String toString()
		{
			// first change SPACE to underscore
			if (key.equals(" "))
				key = "_";
			return key + ", " + time;
		}

		private long time;
		private String key;
	}

	// =============
	// SOFT KEYBOARD
	// =============

	class SoftKeyboard extends JPanel implements ActionListener
	{
		// the following avoids a "warning" with Java 1.5.0 complier (?)
		static final long serialVersionUID = 42L;
		final Font DEFAULT_KEY_FONT = new Font("sansserif", Font.BOLD, 16);

		JButton[] keyButton;
		int mode;

		public SoftKeyboard(Key[] keyArg, ActionListener alArg, int modeArg)
		{
			Key[] key = keyArg;
			keyButton = new JButton[keyArg.length];
			mode = modeArg;

			// determine required dimension for panel
			int width = 0;
			int height = 0;
			for (int i = 0; i < key.length; ++i)
			{
				int wTemp = key[i].x + key[i].w;
				width = wTemp > width ? wTemp : width;

				int hTemp = key[i].y + key[i].h;
				height = hTemp > height ? hTemp : height;
			}
			Dimension d = new Dimension(width, height);

			for (int i = 0; i < keyButton.length; ++i)
			{
				keyButton[i] = new JButton(key[i].getLabel());
				if (mode == METROPOLIS)
				{
					keyButton[i].setUI(new MetropolisButtonUI());
					keyButton[i].setOpaque(false);
				} else
					keyButton[i].setBackground(Color.LIGHT_GRAY);
				keyButton[i].setFont(DEFAULT_KEY_FONT);
				keyButton[i].addActionListener(alArg);
				keyButton[i].setBounds(key[i].getX(), key[i].getY(), key[i].getW(), key[i].getH());
				this.add(keyButton[i]);
			}
			this.setLayout(null);
			this.setPreferredSize(d);
			this.setMaximumSize(d);
		}

		ActionEvent ae;

		public ActionEvent getActionEvent()
		{
			return ae;
		}

		public void setActionEvent(ActionEvent aeArg)
		{
			ae = aeArg;
		}

		public void actionPerformed(ActionEvent ae)
		{
			setActionEvent(ae);
		}
	}

	int w = 50;
	int h = 45;
	Key[] qwertyKeyboard = { new Key("Q", 0 * w, 0, w, h), new Key("W", 1 * w, 0, w, h), new Key("E", 2 * w, 0, w, h),
			new Key("R", 3 * w, 0, w, h), new Key("T", 4 * w, 0, w, h), new Key("Y", 5 * w, 0, w, h),
			new Key("U", 6 * w, 0, w, h), new Key("I", 7 * w, 0, w, h), new Key("O", 8 * w, 0, w, h),
			new Key("P", 9 * w, 0, w, h), new Key("A", (w / 3) + 0 * w, h, w, h),
			new Key("S", (w / 3) + 1 * w, h, w, h), new Key("D", (w / 3) + 2 * w, h, w, h),
			new Key("F", (w / 3) + 3 * w, h, w, h), new Key("G", (w / 3) + 4 * w, h, w, h),
			new Key("H", (w / 3) + 5 * w, h, w, h), new Key("J", (w / 3) + 6 * w, h, w, h),
			new Key("K", (w / 3) + 7 * w, h, w, h), new Key("L", (w / 3) + 8 * w, h, w, h),
			new Key("Z", (2 * w / 3) + 0 * w, 2 * h, w, h), new Key("X", (2 * w / 3) + 1 * w, 2 * h, w, h),
			new Key("C", (2 * w / 3) + 2 * w, 2 * h, w, h), new Key("V", (2 * w / 3) + 3 * w, 2 * h, w, h),
			new Key("B", (2 * w / 3) + 4 * w, 2 * h, w, h), new Key("N", (2 * w / 3) + 5 * w, 2 * h, w, h),
			new Key("M", (2 * w / 3) + 6 * w, 2 * h, w, h), new Key("Space", 2 * w, 3 * h, 6 * w, h),
			new Key("Enter", (w / 3) + 9 * w, 1 * h, 2 * w, h), new Key("Bksp", 10 * w, 0, 2 * w, h) };

	Key[] optiKeyboard = { new Key("Q", 0 * w, 0, w, h), new Key("F", 1 * w, 0, w, h), new Key("U", 2 * w, 0, w, h),
			new Key("M", 3 * w, 0, w, h), new Key("C", 4 * w, 0, w, h), new Key("K", 5 * w, 0, w, h),
			new Key("Z", 6 * w, 0, w, h), new Key("Space", 0 * w, h, 2 * w, h), new Key("O", 2 * w, h, 1 * w, h),
			new Key("T", 3 * w, h, 1 * w, h), new Key("H", 4 * w, h, 1 * w, h), new Key("Space", 5 * w, h, 2 * w, h),
			new Key("B", 0 * w, 2 * h, w, h), new Key("S", 1 * w, 2 * h, w, h), new Key("R", 2 * w, 2 * h, w, h),
			new Key("E", 3 * w, 2 * h, w, h), new Key("A", 4 * w, 2 * h, w, h), new Key("W", 5 * w, 2 * h, w, h),
			new Key("X", 6 * w, 2 * h, w, h), new Key("Space", 0 * w, 3 * h, 2 * w, h),
			new Key("I", 2 * w, 3 * h, 1 * w, h), new Key("N", 3 * w, 3 * h, 1 * w, h),
			new Key("D", 4 * w, 3 * h, 1 * w, h), new Key("Space", 5 * w, 3 * h, 2 * w, h),
			new Key("J", 0 * w, 4 * h, w, h), new Key("P", 1 * w, 4 * h, w, h), new Key("V", 2 * w, 4 * h, w, h),
			new Key("G", 3 * w, 4 * h, w, h), new Key("L", 4 * w, 4 * h, w, h), new Key("Y", 5 * w, 4 * h, w, h),
			new Key("Bksp", 7 * w, 0, 2 * w, h), new Key("Enter", 7 * w, 2 * h, 2 * w, h), };

	Key[] lewisKeyboard = { new Key("Q", 0 * w, 0, w, h), new Key("R", 1 * w, 0, w, h), new Key("W", 2 * w, 0, w, h),
			new Key("X", 3 * w, 0, w, h), new Key("Y", 4 * w, 0, w, h), new Key("Bksp", 5 * w, 0, 2 * w, h),
			new Key("L", 0 * w, 1 * h, w, h), new Key("U", 1 * w, 1 * h, w, h), new Key("A", 2 * w, 1 * h, w, h),
			new Key("O", 3 * w, 1 * h, w, h), new Key("F", 4 * w, 1 * h, w, h), new Key("T", 0 * w, 2 * h, w, h),
			new Key("H", 1 * w, 2 * h, w, h), new Key("E", 2 * w, 2 * h, w, h), new Key("N", 3 * w, 2 * h, w, h),
			new Key("G", 4 * w, 2 * h, w, h), new Key("Enter", 5 * w, 2 * h, 2 * w, h),
			new Key("V", 0 * w, 3 * h, w, h), new Key("D", 1 * w, 3 * h, w, h), new Key("I", 2 * w, 3 * h, w, h),
			new Key("S", 3 * w, 3 * h, w, h), new Key("P", 4 * w, 3 * h, w, h), new Key("B", 0 * w, 4 * h, w, h),
			new Key("C", 1 * w, 4 * h, w, h), new Key("M", 2 * w, 4 * h, w, h), new Key("J", 3 * w, 4 * h, w, h),
			new Key("K", 4 * w, 4 * h, w, h), new Key("Z", 5 * w, 4 * h, w, h), new Key("Space", 0, 5 * h, 6 * w, h) };

	Key[] fitalyKeyboard = { new Key("Z", 0 * w, 0, w, h), new Key("V", 1 * w, 0, w, h), new Key("C", 2 * w, 0, w, h),
			new Key("H", 3 * w, 0, w, h), new Key("W", 4 * w, 0, w, h), new Key("K", 5 * w, 0, w, h),
			new Key("Bksp", 6 * w, 0, 2 * w, h), new Key("F", 0 * w, 1 * h, w, h), new Key("I", 1 * w, 1 * h, w, h),
			new Key("T", 2 * w, 1 * h, w, h), new Key("A", 3 * w, 1 * h, w, h), new Key("L", 4 * w, 1 * h, w, h),
			new Key("Y", 5 * w, 1 * h, w, h), new Key("Space", 0 * w, 2 * h, 2 * w, h),
			new Key("N", 2 * w, 2 * h, w, h), new Key("E", 3 * w, 2 * h, w, h),
			new Key("Space", 4 * w, 2 * h, 2 * w, h), new Key("Enter", 6 * w, 2 * h, 2 * w, h),
			new Key("G", 0 * w, 3 * h, w, h), new Key("D", 1 * w, 3 * h, w, h), new Key("O", 2 * w, 3 * h, w, h),
			new Key("R", 3 * w, 3 * h, w, h), new Key("S", 4 * w, 3 * h, w, h), new Key("B", 5 * w, 3 * h, w, h),
			new Key("Q", 0 * w, 4 * h, w, h), new Key("J", 1 * w, 4 * h, w, h), new Key("U", 2 * w, 4 * h, w, h),
			new Key("M", 3 * w, 4 * h, w, h), new Key("P", 4 * w, 4 * h, w, h), new Key("X", 5 * w, 4 * h, w, h) };

	int mw = 52;
	int mh = 48;
	Key[] metropolisKeyboard = { new Key("J", 2 * mw, 0, mw, mh), new Key("B", 3 * mw, 0, mw + 1, mh),
			new Key("Q", mw / 2 + 0 * mw, 3 * mh / 4, mw, mh + 1), new Key("U", mw / 2 + 1 * mw, 3 * mh / 4, mw, mh),
			new Key("S", mw / 2 + 2 * mw, 3 * mh / 4, mw, mh), new Key("P", mw / 2 + 3 * mw, 3 * mh / 4, mw, mh),
			new Key("C", mw / 2 + 4 * mw, 3 * mh / 4, mw, mh),
			new Key("Bksp", mw / 2 + 5 * mw, 3 * mh / 4, mw + 1, mh + 1),

			new Key("F", 1 * mw, 2 * 3 * mh / 4, mw, mh), new Key("O", 2 * mw, 2 * 3 * mh / 4, mw, mh),
			new Key("T", 3 * mw, 2 * 3 * mh / 4, mw, mh), new Key("H", 4 * mw, 2 * 3 * mh / 4, mw, mh),
			new Key("V", 5 * mw, 2 * 3 * mh / 4, mw + 1, mh),

			new Key("G", mw / 2 + 0 * mw, 3 * 3 * mh / 4, mw, mh),
			new Key("R", mw / 2 + 1 * mw, 3 * 3 * mh / 4, mw, mh),
			new Key("Space", mw / 2 + 2 * mw, 3 * 3 * mh / 4, mw, mh),
			new Key("E", mw / 2 + 3 * mw, 3 * 3 * mh / 4, mw, mh),
			new Key("K", mw / 2 + 4 * mw, 3 * 3 * mh / 4, mw, mh),
			new Key("Enter", mw / 2 + 5 * mw, 3 * 3 * mh / 4, mw + 1, mh + 1),

			new Key("M", 0 * mw, 4 * 3 * mh / 4, mw, mh + 1), new Key("N", 1 * mw, 4 * 3 * mh / 4, mw, mh + 1),
			new Key("I", 2 * mw, 4 * 3 * mh / 4, mw, mh), new Key("A", 3 * mw, 4 * 3 * mh / 4, mw, mh),
			new Key("W", 4 * mw, 4 * 3 * mh / 4, mw, mh), new Key("Z", 5 * mw, 4 * 3 * mh / 4, mw + 1, mh + 1),

			new Key("D", mw / 2 + 1 * mw, 5 * 3 * mh / 4, mw, mh + 1),
			new Key("L", mw / 2 + 2 * mw, 5 * 3 * mh / 4, mw, mh + 1),
			new Key("Y", mw / 2 + 3 * mw, 5 * 3 * mh / 4, mw, mh + 1),
			new Key("X", mw / 2 + 4 * mw, 5 * 3 * mh / 4, mw + 1, mh + 1), };

	// =============================
	// === Metropolis buttons UI ===
	// =============================
	class MetropolisButtonUI extends ButtonUI
	{
		// final Insets I = new Insets(10, 10, 10, 10);
		final Insets I = new Insets(16, 18, 16, 18);

		public void installUI(JComponent c)
		{
			AbstractButton b = (AbstractButton)c;
			BasicButtonListener listener = new BasicButtonListener(b);
			b.addMouseListener(listener);
		}

		public void uninstallUI(JComponent c)
		{
			AbstractButton b = (AbstractButton)c;
			BasicButtonListener listener = new BasicButtonListener(b);
			b.removeMouseListener(listener);
		}

		public Insets getDefaultMargin(AbstractButton ab)
		{
			return I;
		}

		public Insets getInsets(JComponent c)
		{
			return I;
		}

		public Dimension getMaximumSize(JComponent c)
		{
			return this.getPreferredSize(c);
		}

		public Dimension getMinimumSize(JComponent c)
		{
			return this.getPreferredSize(c);
		}

		public Dimension getPreferredSize(JComponent c)
		{
			Graphics g = c.getGraphics();
			FontMetrics fm = g.getFontMetrics();

			Dimension d = new Dimension();
			d.width = fm.getAscent() + I.left + I.right;
			d.height = fm.getAscent() + I.top + I.bottom;
			return d;
		}

		public void paint(Graphics g, JComponent c)
		{
			Graphics2D g2 = (Graphics2D)g;

			AbstractButton b = (AbstractButton)c;
			ButtonModel bm = b.getModel();

			FontMetrics fm = g2.getFontMetrics();

			// define some colors for drawing
			final Color BLUE_LIGHT = Color.gray;
			final Color BLUE_DARK = Color.lightGray;

			Color backgroundColor;

			if (bm.isArmed())
			{
				backgroundColor = BLUE_LIGHT;
				b.setCursor(new Cursor(Cursor.HAND_CURSOR));
			} else if (bm.isPressed())
				backgroundColor = BLUE_LIGHT;
			else
				backgroundColor = BLUE_DARK; // normal!

			Dimension d = c.getPreferredSize();
			int x = d.width - 1; // x-coordinate of right edge
			int y = d.height - 1; // y-coordinate of bottom edge

			int[] xx = { 2, 4, 4, 2, 0, 0 };
			int[] yy = { 0, 1, 3, 4, 3, 1 };
			for (int i = 0; i < xx.length; ++i)
			{
				xx[i] = xx[i] * (x / 4);
				yy[i] = yy[i] * (y / 4);
			}
			// xx[1] -= 1; // attempt to fix right edge line-missing problem
			// xx[2] -= 1;
			Polygon p = new Polygon(xx, yy, xx.length);

			// g2.setColor(BLUE);

			g2.setColor(backgroundColor);
			g2.fillPolygon(p);

			g2.setColor(Color.black);
			g2.drawPolygon(p);

			// paint foreground
			g2.setColor(Color.black);
			Font f = fm.getFont();
			g2.setFont(new Font(f.getName(), Font.BOLD, f.getSize()));
			String s = ((JButton)c).getText();
			int width = fm.stringWidth(s);
			int height = fm.getAscent();
			x = (d.width - width) / 2;
			y = (d.height - 2 + height - 2) / 2;
			g2.drawString(s, x, y);
		}

		/**
		 * Returns true if x,y is within the rounded rectangle representing this button.
		 */
		public boolean contains(JComponent c, int x, int y)
		{
			int w = c.getWidth();
			int h = c.getHeight();

			int[] xx = { 2, 4, 4, 2, 0, 0, 2 };
			int[] yy = { 0, 1, 3, 4, 3, 1, 0 };
			for (int i = 0; i < xx.length; ++i)
			{
				xx[i] = xx[i] * (w / 4);
				yy[i] = yy[i] * (h / 4);
			}
			Polygon p = new Polygon(xx, yy, xx.length);
			return p.contains(x, y);
		}
	}; // ***** end MetropolisButtonUI *****

	class Key
	{
		String label;
		int x, y, w, h; // x- and y-position, width, height

		Key(String labelArg, int xArg, int yArg, int wArg, int hArg)
		{
			label = labelArg;
			x = xArg;
			y = yArg;
			w = wArg;
			h = hArg;
		}

		String getLabel()
		{
			if (label.equals("Space"))
				return " ";
			else
				return label;
		};

		int getX()
		{
			return x;
		};

		int getY()
		{
			return y;
		};

		int getW()
		{
			return w;
		};

		int getH()
		{
			return h;
		};
	};
}